# Copyright (c) 2022 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
import glob
import itertools
import os

from flask import render_template

from signac_dashboard.module import Module


class VideoViewer(Module):
    """Displays videos that match a glob.

    The VideoViewer module displays videos using an HTML ``<video>`` tag. The
    module defaults to showing all videos of MP4 or M4V types in the job or
    project directory. A filename or glob can be defined to select specific
    filenames, which may be of any format supported by your browser with the
    ``<video>`` tag. Each matching file yields a card.

    A "poster" can be defined, which shows a
    thumbnail with that filename before the video is started. Videos do not
    preload by default since file sizes can be large and there may be many
    videos on a page. To enable preloading, use the argument ``preload='auto'``
    or ``preload='metadata'``.

    Multiple VideoViewer modules can be defined
    with different filenames or globs to enable/disable cards individually.

    :Example:

    .. code-block:: python

        from signac_dashboard.modules import VideoViewer
        video_mod = VideoViewer()  # Shows all MP4/M4V videos
        video_mod = VideoViewer(name='Cool Science Video',
                                context="ProjectContext",
                                video_globs=['cool_science.mp4'],
                                poster='cool_science_thumbnail.jpg',
                                preload='none')

    :param context: Supports :code:`'JobContext'` and :code:`'ProjectContext'`.
    :type context: str
    :param video_globs: A list of glob expressions or exact filenames,
        relative to the job or project directory, to be displayed
        (default: :code:`['*.mp4', '*.m4v']`).
    :type video_globs: list
    :param preload: Option for preloading videos, one of :code:`'auto'`,
        :code:`'metadata'`, or :code:`'none'` (default: :code:`'none'`).
    :type preload: str
    :param poster: A path in the job directory or project directory for a
        poster image to be shown before a video begins playback (default: :code:`None`).
    :type poster: str
    :type sort_key: callable
    :param sort_key: Key to sort the video files, passed internally to :code:`sorted`.

    """

    _supported_contexts = {"JobContext", "ProjectContext"}

    def __init__(
        self,
        name="Video Viewer",
        context="JobContext",
        template="cards/video_viewer.html",
        video_globs=("*.mp4", "*.m4v"),
        preload="none",  # auto|metadata|none
        poster=None,
        sort_key=None,
        **kwargs,
    ):
        super().__init__(
            name=name,
            context=context,
            template=template,
            **kwargs,
        )
        self.video_globs = video_globs
        self.preload = preload
        self.poster = poster
        self.sort_key = sort_key

    def get_cards(self, job_or_project):
        if self.context == "JobContext":
            jobid = job_or_project._id
        elif self.context == "ProjectContext":
            jobid = None

        def make_card(filename):
            if not job_or_project.isfile(filename):
                raise FileNotFoundError(
                    "The filename {} could not be found "
                    "for {}.".format(
                        filename, "the project" if jobid is None else f"job {jobid}"
                    )
                )
            return {
                "name": self.name + ": " + filename,
                "content": render_template(
                    self.template,
                    jobid=jobid,
                    poster=self.poster
                    if self.poster and job_or_project.isfile(self.poster)
                    else None,
                    preload=self.preload,
                    filename=filename,
                ),
            }

        video_globs = [
            glob.iglob(job_or_project.fn(video_glob)) for video_glob in self.video_globs
        ]
        video_files = itertools.chain(*video_globs)
        video_files = sorted(video_files, key=self.sort_key)
        for filepath in video_files:
            yield make_card(os.path.relpath(filepath, job_or_project.fn("")))



# Copyright (c) 2022 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.

from flask import (
    abort,
    flash,
    g,
    redirect,
    render_template,
    request,
    send_from_directory,
    session,
    url_for,
)


def home(dashboard):
    return redirect(url_for("project_info"))


def search(dashboard):
    g.query = request.args.get("q", None)
    jobs = []
    try:
        if request.method != "GET":
            # Someday we may support search via POST, returning json
            raise NotImplementedError("Unsupported search method.")
        jobs = dashboard._job_search(g.query)
        if not jobs:
            flash("No jobs found for the provided query.", "warning")
    except Exception as error:
        return dashboard._render_error(error)
    else:
        g.pagination = dashboard._setup_pagination(jobs)
        g.jobs = dashboard._get_job_details(g.pagination.paginate(jobs))
        g.title = f"Search: {g.query}"
        g.subtitle = g.pagination.item_counts()
        return dashboard._render_job_view(default_view="list")


def jobs_list(dashboard):
    jobs = dashboard._get_all_jobs()
    g.pagination = dashboard._setup_pagination(jobs)
    if not jobs:
        flash("No jobs found.", "warning")
    g.jobs = dashboard._get_job_details(g.pagination.paginate(jobs))
    g.title = "signac-dashboard: Jobs"
    g.subtitle = g.pagination.item_counts()
    return dashboard._render_job_view(default_view="list")


def project_info(dashboard):
    g.project = dashboard.project
    g.title = "signac-dashboard"
    num_jobs = len(dashboard.project)
    g.subtitle = f"{num_jobs} jobs"
    return dashboard._render_project_view()


def show_job(dashboard, jobid):
    try:
        job = dashboard.project.open_job(id=jobid)
    except KeyError:
        abort(404, "The job id requested could not be found.")
    else:
        g.jobs = dashboard._get_job_details([job])
        g.title = g.jobs[0]["title"]
        g.subtitle = g.jobs[0]["subtitle"]
        return dashboard._render_job_view(default_view="grid")


def get_file(dashboard, filename, jobid=None):
    if jobid is not None:
        try:
            job_or_project = dashboard.project.open_job(id=jobid)
        except KeyError:
            abort(404, "The job id requested could not be found.")
        except LookupError:
            dashboard.project.find_jobs()
            abort(404, "Multiple jobs match the requested job id.")
    else:
        job_or_project = dashboard.project
    if job_or_project.isfile(filename):
        directory = job_or_project.fn("")
        mimetype = None
        max_age = 0
        download_name = request.args.get("download_name", filename)
        return send_from_directory(
            directory=directory,
            path=filename,
            mimetype=mimetype,
            max_age=max_age,
            conditional=True,
            download_name=download_name,
        )
    else:
        abort(404, "The file requested does not exist.")


def change_modules(dashboard):
    enabled_module_indices = session.get(
        "enabled_module_indices", dashboard._setup_enabled_module_indices()
    )
    enabled_module_indices = {
        k: set(v) for k, v in enabled_module_indices.items()
    }  # remove duplicates
    context = session.get("context", "JobContext")

    for i, module in enumerate(dashboard._modules_by_context[context]):
        if request.form.get(f"modules[{i}]") == "on":
            enabled_module_indices[context].add(i)
        else:
            enabled_module_indices[context].discard(i)

    session["enabled_module_indices"] = {
        k: list(v) for k, v in enabled_module_indices.items()
    }
    return redirect(request.form.get("redirect", url_for("home")))


def settings(dashboard):
    g.active_page = "settings"
    return render_template("settings.html")


def page_not_found(dashboard, error):
    return dashboard._render_error(str(error))



# Copyright (c) 2022 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
import glob
import itertools
import os

from flask import render_template

from signac_dashboard.module import Module


class ImageViewer(Module):
    """Displays images that match a glob.

    The ImageViewer module can display images in any format that works with a standard
    ``<img>`` tag. The module defaults to showing all images of PNG, JPG, GIF, and SVG
    types in the job or project directory. A filename or glob can be
    defined to select specific filenames. Each matching file yields a card.

    Multiple ImageViewer modules can be defined with different filenames or
    globs to enable/disable cards for each image or image group.

    :Example:

    .. code-block:: python

        from signac_dashboard.modules import ImageViewer
        img_mod = ImageViewer()  # Show all PNG/JPG/GIF/SVG images
        img_mod = ImageViewer(name='Bond Order Diagram', img_globs=['bod.png'])
        img_mod = ImageViewer(context="ProjectContext",
                              img_globs=['/gallery/*.png'])  # search subdirectory of project path

    :param context: Supports :code:`'JobContext'` and :code:`'ProjectContext'`.
    :type context: str
    :param img_globs: A list of glob expressions or exact filenames,
        relative to the job or project directory, to be displayed
        (default: :code:`['*.png', '*.jpg', '*.gif', '*.svg']`).
    :type img_globs: list
    :type sort_key: callable
    :param sort_key: Key to sort the image files, passed internally to :code:`sorted`.

    """

    _supported_contexts = {"JobContext", "ProjectContext"}

    def __init__(
        self,
        name="Image Viewer",
        context="JobContext",
        template="cards/image_viewer.html",
        img_globs=("*.png", "*.jpg", "*.gif", "*.svg"),
        sort_key=None,
        **kwargs,
    ):
        super().__init__(
            name=name,
            context=context,
            template=template,
            **kwargs,
        )
        self.img_globs = img_globs
        self.sort_key = sort_key

    def get_cards(self, job_or_project):
        if self.context == "JobContext":
            jobid = job_or_project._id
            modal_label = job_or_project._id
        elif self.context == "ProjectContext":
            jobid = None
            modal_label = "project"

        def make_card(filename):
            return {
                "name": self.name + ": " + filename,
                "content": render_template(
                    self.template,
                    modal_label=modal_label,
                    jobid=jobid,
                    filename=filename,
                ),
            }

        image_globs = [
            glob.iglob(job_or_project.fn(image_glob)) for image_glob in self.img_globs
        ]
        image_files = itertools.chain(*image_globs)
        image_files = sorted(image_files, key=self.sort_key)
        for filepath in image_files:
            yield make_card(os.path.relpath(filepath, job_or_project.fn("")))



# Copyright (c) 2022 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
import json
import re
import shutil
import tempfile
import unittest
from urllib.parse import quote as urlquote

from signac import init_project

import signac_dashboard.modules
from signac_dashboard import Dashboard


class DashboardTestCase(unittest.TestCase):
    def get_response(self, query):
        rv = self.test_client.get(query, follow_redirects=True)
        return str(rv.get_data())

    def setUp(self):
        self._tmp_dir = tempfile.mkdtemp()
        self.project = init_project(self._tmp_dir)
        # Set up some fake jobs
        for a in range(3):
            for b in range(2):
                job = self.project.open_job({"a": a, "b": b})
                with job:
                    job.document["sum"] = a + b
        self.config = {"ACCESS_TOKEN": "test"}
        self.modules = []
        self.dashboard = Dashboard(
            config=self.config, project=self.project, modules=self.modules
        )
        self.test_client = self.dashboard.app.test_client()
        self.addCleanup(shutil.rmtree, self._tmp_dir)

        # Test logged out content
        response = self.get_response("/")
        assert "Login required" in response

        response = self.get_response("/jobs/7f9fb369851609ce9cb91404549393f3")
        assert "Login required" in response

        response = self.get_response("/login?token=error")
        assert "Login required" in response
        assert "Incorrect token" in response

        # login
        self.test_client.get("/login?token=test", follow_redirects=True)

    def test_get_project(self):
        rv = self.test_client.get("/project/", follow_redirects=True)
        response = str(rv.get_data())
        assert "signac-dashboard" in response

    def test_get_jobs(self):
        rv = self.test_client.get("/jobs/", follow_redirects=True)
        response = str(rv.get_data())
        assert "signac-dashboard: Jobs" in response

    def test_job_count(self):
        rv = self.test_client.get("/jobs/", follow_redirects=True)
        response = str(rv.get_data())
        assert f"{len(self.project)} jobs" in response

    def test_sp_search(self):
        dictquery = {"a": 0}
        true_num_jobs = len(list(self.project.find_jobs(dictquery)))
        query = urlquote(json.dumps(dictquery))
        rv = self.test_client.get(f"/search?q={query}", follow_redirects=True)
        response = str(rv.get_data())
        assert f"{true_num_jobs} jobs" in response

    def test_doc_search(self):
        dictquery = {"doc.sum": 1}
        true_num_jobs = len(list(self.project.find_jobs(dictquery)))
        query = urlquote(json.dumps(dictquery))
        rv = self.test_client.get(f"/search?q={query}", follow_redirects=True)
        response = str(rv.get_data())
        assert f"{true_num_jobs} jobs" in response

    def test_allow_where_search(self):
        dictquery = {"doc.sum": 1}
        true_num_jobs = len(list(self.project.find_jobs(dictquery)))
        query = urlquote('doc.sum.$where "lambda x: x == 1"')

        self.dashboard.config["ALLOW_WHERE"] = False
        rv = self.test_client.get(f"/search?q={query}", follow_redirects=True)
        response = str(rv.get_data())
        assert "ALLOW_WHERE must be enabled for this query." in response

        self.dashboard.config["ALLOW_WHERE"] = True
        rv = self.test_client.get(f"/search?q={query}", follow_redirects=True)
        response = str(rv.get_data())
        assert f"{true_num_jobs} jobs" in response

    def test_update_cache(self):
        rv = self.test_client.get("/jobs", follow_redirects=True)
        response = str(rv.get_data())
        assert f"{len(self.project)} jobs" in response

        # Create a new job. Because the project has been cached, the response
        # will be wrong until the cache is cleared.
        self.project.open_job({"a": "test-cache"}).init()
        rv = self.test_client.get("/jobs", follow_redirects=True)
        response = str(rv.get_data())
        assert f"{len(self.project)} jobs" not in response

        # Clear cache and try again.
        self.dashboard.update_cache()
        rv = self.test_client.get("/jobs", follow_redirects=True)
        response = str(rv.get_data())
        assert f"{len(self.project)} jobs" in response

    def test_no_view_single_job(self):
        """Make sure View panel is not shown when on a single job page."""
        response = self.get_response("/jobs/7f9fb369851609ce9cb91404549393f3")
        assert "Views" not in response

    def test_logout(self):
        response = self.get_response("/logout")
        if self.dashboard.config.get("ACCESS_TOKEN") is not None:
            assert "Login required" in response


class NoModulesTestCase(DashboardTestCase):
    """Test the inherited tests and cases without any modules."""

    def test_job_sidebar(self):
        response = self.get_response("/jobs/?view=grid")
        assert "No modules." in response

    def test_project_sidebar(self):
        response = self.get_response("/project/")
        assert "No modules." in response
        assert "Views" not in response


class AllModulesTestCase(DashboardTestCase):
    """Add all modules and contexts and test again."""

    def setUp(self):
        self._tmp_dir = tempfile.mkdtemp()
        self.project = init_project(self._tmp_dir)
        # Set up some fake jobs
        for a in range(3):
            for b in range(2):
                job = self.project.open_job({"a": a, "b": b})
                with job:
                    job.document["sum"] = a + b
        self.config = {"ACCESS_TOKEN": None}
        modules = []
        for m in signac_dashboard.modules.__all__:
            module = getattr(signac_dashboard.modules, m)
            for c in module._supported_contexts:
                modules.append(module(context=c))
                with self.assertRaises(RuntimeError):
                    module(context="BadContext")
        self.modules = modules
        self.dashboard = Dashboard(
            config=self.config, project=self.project, modules=self.modules
        )
        self.test_client = self.dashboard.app.test_client()
        self.addCleanup(shutil.rmtree, self._tmp_dir)

    def test_login_with_None_token(self):
        rv = self.test_client.get("/login", follow_redirects=True)
        response = str(rv.get_data())
        assert "signac-dashboard" in response

    def test_module_visible_mobile(self):
        response = self.get_response("/jobs/?view=grid")
        # Check for two instances of Modules header
        pattern = re.compile("Modules</h")
        module_headers = re.findall(pattern, response)
        assert len(module_headers) == 2

    def test_module_selector(self):
        project_response = self.get_response("/project/")
        job_response = self.get_response("/jobs/?view=grid")
        for m in self.modules:
            print(f"Checking for {m.name} in {m.context}.")
            if m.context == "ProjectContext":
                assert m.name in project_response
            elif m.context == "JobContext":
                assert m.name in job_response

    def test_enabled_module_indices_project_session(self):
        """Ensure that the message is not displayed when modules are actually enabled."""
        project_response = self.get_response("/project/")
        assert "No modules for the ProjectContext are enabled." not in project_response

    def test_enabled_module_indices_job_session(self):
        """Ensure that the message is not displayed when modules are actually enabled."""
        job_response = self.get_response("/jobs/?view=grid")
        assert "No modules for the JobContext are enabled." not in job_response

    def test_navigator_module(self):
        """Look for the next and previous values in a table on the page."""
        response = self.get_response("/jobs/017d53deb17a290d8b0d2ae02fa8bd9d")
        # next job for a = 2
        assert '<a href="/jobs/fb4e5868559e719f0c5826de08023281"' in response
        # next job for b = 1
        assert '<a href="/jobs/386b19932c82f3f9749dd6611e846293"' in response
        assert "disabled>min</div>" in response  # no previous job for b


if __name__ == "__main__":
    unittest.main()



# Copyright (c) 2022 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.

import flask_login
from markupsafe import escape
from werkzeug.utils import cached_property, import_string


def ellipsis_truncate_middle(val, length=60):
    string = str(val)
    half = int(length / 2)
    if len(string) < length:
        return str(escape(string))
    else:
        return str(escape(string[:half])) + "&hellip;" + str(escape(string[-half:]))


def ellipsis_truncate_end(val, length=60):
    """Escape and truncate val to length with html ellipsis at end."""
    string = str(val)
    if len(string) < length:
        return str(escape(string))
    else:
        return str(escape(string[:length])) + "&hellip;"


def abbr_value(val, max_chars):
    if len(str(val)) > max_chars:
        link_string = str(ellipsis_truncate_end(val, length=max_chars))
        return f'<abbr title="{escape(val)}">{link_string}</abbr>'
    else:
        return str(escape(val))


def escape_truncated_values(data, max_chars):
    """Truncate values in a dict to a maximum number of characters."""
    if max_chars is not None and int(max_chars) > 0:
        for key in data:
            if len(str(data[key])) > max_chars:
                data[key] = (
                    str(escape(ellipsis_truncate_middle(data[key], length=max_chars)))
                    + " <em>[Truncated]</em>"
                )
    else:
        for key in data:
            data[key] = escape(data[key])
    return data


class LazyView:
    # See https://flask.palletsprojects.com/en/2.1.x/patterns/lazyloading/
    def __init__(self, dashboard, import_name):
        self.__module__, self.__name__ = import_name.rsplit(".", 1)
        self.import_name = import_name
        self.dashboard = dashboard

    @cached_property
    def view(self):
        return import_string(self.import_name)

    def __call__(self, *args, **kwargs):
        # Protect routes added in the format self.add_url("views.home", ["/"])
        if not flask_login.current_user.is_authenticated:
            return self.dashboard.login_manager.unauthorized(), 401

        kwargs.update({"dashboard": self.dashboard})
        return self.view(*args, **kwargs)



# Copyright (c) 2022 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
import sys

import signac

from . import Dashboard
from .modules import DocumentList, ImageViewer, StatepointList


def main():
    try:
        project = signac.get_project()
    except LookupError:
        print("No signac project could be found in the current directory.")
        sys.exit(1)

    # Initialize a new Dashboard using essential modules
    modules = [StatepointList(), DocumentList(), ImageViewer()]
    Dashboard(modules=modules, project=project).main()


if __name__ == "__main__":
    main()



# Copyright (c) 2022 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
import logging

from flask import render_template

from signac_dashboard.module import Module

logger = logging.getLogger(__name__)


class FlowStatus(Module):
    """Show job labels from a :py:class:`flow.FlowProject`.

    This module displays a card with labels from
    :py:meth:`flow.FlowProject.labels`. The user must provide an instance of
    :py:class:`flow.FlowProject` to the dashboard constructor. Example:

    .. code-block:: python

        from project import Project  # Project is a FlowProject with labels
        from signac_dashboard import Dashboard

        if __name__ == '__main__':
            Dashboard(project=Project()).main()

    :param context: Supports :code:`'JobContext'`.
    :type context: str
    """

    _supported_contexts = {"JobContext"}

    def __init__(
        self,
        name="Flow Status",
        context="JobContext",
        template="cards/flow_status.html",
        project_module="project",
        project_class="Project",
        **kwargs,
    ):
        super().__init__(
            name=name,
            context=context,
            template=template,
            **kwargs,
        )

    def register(self, dashboard):
        self.project = dashboard.project
        if not hasattr(self.project, "labels"):
            logger.warning(
                "The provided signac Project cannot provide labels. "
                "Try providing a FlowProject to the Dashboard's "
                "project argument."
            )

    def get_cards(self, job):
        try:
            labels = self.project.labels(job)
        except AttributeError:
            labels = ("Error: Project cannot provide labels.",)
        return [
            {
                "name": self.name,
                "content": render_template(self.template, labels=labels),
            }
        ]



# Copyright (c) 2022 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
import flask_login
from flask import abort, render_template, request
from jinja2.exceptions import TemplateNotFound

from signac_dashboard.module import Module


class Notes(Module):
    """Displays a text box that is synced to the job document.

    The contents of the text box are saved to :code:`job.document['notes']`.
    The Notes module can be used to annotate a large data space with tags or
    human-readable descriptions for post-processing, parsing, or searching.

    :param context: Supports :code:`'JobContext'`.
    :type context: str
    :param key: Document key to display and update (default: :code:`'notes'`).
    :type key: str
    """

    _supported_contexts = {"JobContext"}

    def __init__(
        self,
        name="Notes",
        context="JobContext",
        template="cards/notes.html",
        key="notes",
        **kwargs,
    ):
        super().__init__(
            name=name,
            context=context,
            template=template,
            **kwargs,
        )
        self.key = key

    def get_cards(self, job):
        note_text = job.document.get(self.key, "")
        return [
            {
                "name": self.name,
                "content": render_template(
                    self.template, note_text=note_text, jobid=job._id
                ),
            }
        ]

    def register(self, dashboard):
        # Register routes
        @dashboard.app.route("/module/notes/update", methods=["POST"])
        @flask_login.login_required
        def notes_update():
            note_text = request.form.get("note_text")
            jobid = request.form.get("jobid")
            job = dashboard.project.open_job(id=jobid)
            job.document[self.key] = note_text
            return "Saved."

        @dashboard.app.route("/module/notes/<path:filename>")
        @flask_login.login_required
        def notes_asset(filename):
            path = f"notes/{filename}"
            try:
                return render_template(path)
            except TemplateNotFound:
                abort(404, "The file requested does not exist.")

        # Register assets
        assets = ["js/notes.js"]
        for asset_file in assets:
            dashboard.register_module_asset(
                {
                    "file": f"templates/notes/{asset_file}",
                    "url": f"/module/notes/{asset_file}",
                }
            )



# Copyright (c) 2022 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
from ast import literal_eval
from collections import OrderedDict

import flask_login
from flask import abort, render_template, request
from jinja2.exceptions import TemplateNotFound
from markupsafe import escape

from signac_dashboard.module import Module


class DocumentEditor(Module):
    """Provides an interface to edit the job document.

    This module shows keys in the job document with a form that allows users
    to edit their contents. When saving, the edited strings are parsed into
    JSON-compatible Python data structures (e.g., :py:class:`list` and
    :py:class:`dict`). Job document keys beginning with an underscore
    :code:`_` are treated as private and are not displayed.

    :param context: Supports :code:`'JobContext'`.
    :type context: str
    """

    _supported_contexts = {"JobContext"}

    def __init__(
        self,
        name="Document Editor",
        context="JobContext",
        template="cards/document_editor.html",
        **kwargs,
    ):
        super().__init__(
            name=name,
            context=context,
            template=template,
            **kwargs,
        )

    def get_cards(self, job):
        doc = OrderedDict(sorted(job.document.items(), key=lambda t: t[0]))

        for key in doc:
            if key.startswith("_"):
                # Don't allow users to edit "private" keys that begin with _
                del doc[key]
            else:
                doc[key] = escape(repr(doc[key]))
        return [
            {
                "name": self.name,
                "content": render_template(self.template, document=doc, jobid=job._id),
            }
        ]

    def register(self, dashboard):
        # Register routes
        @dashboard.app.route("/module/document_editor/update", methods=["POST"])
        @flask_login.login_required
        def document_editor_update():
            jobid = request.form.get("jobid")
            job = dashboard.project.open_job(id=jobid)
            for key, value in request.form.items():
                if key.startswith("doc:"):
                    key = key[4:]
                    try:
                        job.doc[key] = literal_eval(value)
                    except (SyntaxError, TypeError, ValueError) as e:
                        return (
                            f"Error in key <strong>{key}</strong>: {e}",
                            422,
                        )
            return "Saved."

        @dashboard.app.route("/module/document_editor/<path:filename>")
        @flask_login.login_required
        def document_editor_asset(filename):
            path = f"document_editor/{filename}"
            try:
                return render_template(path)
            except TemplateNotFound:
                abort(404, "The file requested does not exist.")

        # Register assets
        assets = ["js/document_editor.js"]
        for assetfile in assets:
            dashboard.register_module_asset(
                {
                    "file": f"templates/document_editor/{assetfile}",
                    "url": f"/module/document_editor/{assetfile}",
                }
            )



# Copyright (c) 2022 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
from collections import OrderedDict

from flask import render_template

from signac_dashboard.module import Module
from signac_dashboard.util import escape_truncated_values


class DocumentList(Module):
    """Displays the job or project document.

    Long values can be optionally truncated.

    :param context: Supports :code:`'JobContext'` and :code:`'ProjectContext'`.
    :type context: str
    :param max_chars: Truncation length for document values (default:
        :code:`None`).
    :type max_chars: int
    """

    _supported_contexts = {"JobContext", "ProjectContext"}

    def __init__(
        self,
        name=None,
        context="JobContext",
        template="cards/escaped_dict_display.html",
        max_chars=None,
        **kwargs,
    ):
        # Set name based on context
        if name is None:
            if context == "JobContext":
                name = "Job Document"
            elif context == "ProjectContext":
                name = "Project Document"

        super().__init__(
            name=name,
            context=context,
            template=template,
            **kwargs,
        )
        self.max_chars = max_chars

    def get_cards(self, job_or_project):
        doc = OrderedDict(sorted(job_or_project.document.items(), key=lambda t: t[0]))

        # We manually escape the document's contents since the field is marked
        # "safe" in the Jinja template. This is necessary because we added
        # custom HTML for "[Truncated]" fields
        doc = escape_truncated_values(doc, self.max_chars)

        return [
            {
                "name": self.name,
                "content": render_template(self.template, escaped_dict=doc),
            }
        ]



from .document_editor import DocumentEditor
from .document_list import DocumentList
from .file_list import FileList
from .flow_status import FlowStatus
from .image_viewer import ImageViewer
from .navigator import Navigator
from .notes import Notes
from .schema import Schema
from .statepoint_list import StatepointList
from .text_display import TextDisplay
from .video_viewer import VideoViewer

__all__ = [
    "DocumentEditor",
    "DocumentList",
    "FileList",
    "FlowStatus",
    "ImageViewer",
    "Navigator",
    "Notes",
    "Schema",
    "StatepointList",
    "TextDisplay",
    "VideoViewer",
]



# Copyright (c) 2022 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.

import argparse
import inspect
import json
import logging
import os
import secrets
import shlex
import sys
import warnings
from functools import lru_cache
from itertools import groupby
from numbers import Real

import flask_login
import jinja2
import natsort
import signac
from flask import Flask, flash, g, redirect, render_template, request, session, url_for
from flask_assets import Bundle, Environment
from flask_turbolinks import turbolinks
from watchdog.events import FileSystemEventHandler
from watchdog.observers import Observer
from werkzeug.urls import url_encode

from .pagination import Pagination
from .util import LazyView
from .version import __version__

logger = logging.getLogger(__name__)


class _FileSystemEventHandler(FileSystemEventHandler):
    def __init__(self, dashboard):
        self.dashboard = dashboard

    def on_modified(self, event):
        if os.path.realpath(event.src_path) == os.path.realpath(
            self.dashboard.project.workspace
        ):
            self.dashboard.update_cache()


class User(flask_login.UserMixin):
    """User class for flask_login.

    The default implementation of `UserMixin` assumes that all instantiated User
    classes are active and logged in.
    """

    def __init__(self, id):
        self.id = id


class Dashboard:
    """A dashboard application to display a :py:class:`signac.Project`.

    The Dashboard class is designed to be used as a base class for a child
    class such as :code:`MyDashboard` which can be customized and launched via
    its command line interface (CLI). The CLI is invoked by calling
    :py:meth:`.main` on an instance of this class.

    **Configuration options:** The :code:`config` dictionary recognizes the
    following options:

    - **HOST**: Sets binding address (default: localhost).
    - **PORT**: Sets port to listen on (default: 8888).
    - **DEBUG**: Enables debug mode if :code:`True` (default: :code:`False`).
    - **PROFILE**: Enables the profiler
      :py:class:`werkzeug.middleware.profiler.ProfilerMiddleware` if
      :code:`True` (default: :code:`False`).
    - **PER_PAGE**: Maximum number of jobs to show per page
      (default: 24).
    - **CARDS_PER_ROW**: Cards to show per row in the desktop view. Must be a
      factor of 12 (default: 3).
    - **ACCESS_TOKEN**: The access token required to login to the dashboard.
      Set to :code:`None` to disable authentication (not recommended on multi-user
      systems).
    - **SECRET_KEY**: This must be specified to run via WSGI with multiple
      workers, so that sessions remain intact. See the
      `Flask docs <http://flask.pocoo.org/docs/1.0/config/#SECRET_KEY>`_
      for more information.
    - **ALLOW_WHERE**: If True, search queries can include :code:`$where`
      statements, which potentially allows arbitrary code execution from user
      input. *Caution:* This should only be enabled in trusted environments,
      never on a publicly-accessible server (default: :code:`False`).

    :param config: Configuration dictionary (default: :code:`{}`).
    :type config: dict
    :param project: signac project (default: :code:`None`, autodetected).
    :type project: :py:class:`signac.Project`
    :param modules: List of :py:class:`~.Module` instances to display.
    :type modules: list
    """

    def __init__(self, config={}, project=None, modules=[]):
        if project is None:
            self.project = signac.get_project()
        else:
            self.project = project
        self.project.update_cache()

        self.config = config
        self.modules = modules

        self.event_handler = _FileSystemEventHandler(self)
        self.observer = Observer()
        self.observer.schedule(self.event_handler, self.project.workspace)

        # Prepare this dashboard instance to run.

        # Set configuration defaults
        self.config.setdefault("HOST", "localhost")
        self.config.setdefault("DEBUG", False)
        self.config.setdefault("PORT", 8888)
        self.config.setdefault("PAGINATION", True)
        self.config.setdefault("PER_PAGE", 24)
        self.config.setdefault("CARDS_PER_ROW", 3)
        if 12 % self.config["CARDS_PER_ROW"] != 0:
            raise ValueError(
                "The value of CARDS_PER_ROW must be a factor of 12. Got "
                f"{self.config['CARDS_PER_ROW']}."
            )

        self.config.setdefault("ACCESS_TOKEN", secrets.token_hex(24))

        # Create and configure the Flask application
        self.app = self._create_app(self.config)

        # Initialize the login manager
        self.login_manager = flask_login.LoginManager()
        self.login_manager.init_app(self.app)

        @self.login_manager.user_loader
        def user_loader(identifier):
            if self.config["ACCESS_TOKEN"] is None:
                return User("None")

            if secrets.compare_digest(identifier, self.config["ACCESS_TOKEN"]):
                return User(identifier)

            return None

        @self.login_manager.request_loader
        def load_user_from_request(request):
            if self.config["ACCESS_TOKEN"] is None:
                return User("None")

            return None

        # Add assets and routes
        self.assets = self._create_assets()
        self._register_routes()

        # Add module assets and routes
        self._module_assets = []
        for module in self.modules:
            try:
                module.register(self)
            except Exception as e:
                logger.error(f"Error while registering {module.name} module: {e}")
                logger.error(f"Removing module {module.name} from dashboard.")
                self.modules.remove(module)

        # Group modules to track enabled state
        def keyfunc(module):
            return module.context

        grouped = groupby(sorted(self.modules, key=keyfunc), key=keyfunc)
        modules_by_context = {}
        for context_key, context_group in grouped:
            modules_by_context[context_key] = [m for m in context_group]
        self._modules_by_context = modules_by_context

    def _create_app(self, config={}):
        """Create a Flask application.

        :param config: Dictionary of configuration parameters.
        """
        app = Flask("signac-dashboard")
        app.config.update(
            {
                "SECRET_KEY": os.urandom(24),
                "SEND_FILE_MAX_AGE_DEFAULT": 300,  # Cache control for static files
            }
        )

        # Load the provided config
        app.config.update(config)

        # Enable profiling
        if app.config.get("PROFILE"):
            logger.warning("Application profiling is enabled.")
            from werkzeug.contrib.profiler import ProfilerMiddleware

            app.wsgi_app = ProfilerMiddleware(app.wsgi_app, restrictions=[10])

        # Set up default signac-dashboard static and template paths
        signac_dashboard_path = os.path.dirname(__file__)
        app.static_folder = signac_dashboard_path + "/static"
        app.template_folder = signac_dashboard_path + "/templates"

        # Set up custom template paths
        # The paths in DASHBOARD_PATHS give the preferred order of template
        # loading
        loader_list = []
        for dash_path in list(app.config.get("DASHBOARD_PATHS", [])):
            logger.warning(f"Adding '{dash_path}' to dashboard paths.")
            loader_list.append(jinja2.FileSystemLoader(dash_path + "/templates"))

        # The default loader goes last and is overridden by any custom paths
        loader_list.append(app.jinja_loader)

        app.jinja_loader = jinja2.ChoiceLoader(loader_list)

        turbolinks(app)

        return app

    def _create_assets(self):
        """Add assets for inclusion in the dashboard HTML."""
        assets = Environment(self.app)
        # jQuery is served as a standalone file
        jquery = Bundle("js/jquery-*.min.js", output="gen/jquery.min.js")
        # JavaScript is combined into one file and minified
        js_all = Bundle("js/js_all/*.js", filters="jsmin", output="gen/app.min.js")
        # SCSS (Sassy CSS) is compiled to CSS
        scss_all = Bundle("scss/app.scss", filters="libsass", output="gen/app.css")
        assets.register("jquery", jquery)
        assets.register("js_all", js_all)
        assets.register("scss_all", scss_all)
        return assets

    def register_module_asset(self, asset):
        """Register an asset required by a dashboard module.

        Some modules require special scripts or stylesheets, like the
        :py:class:`signac_dashboard.modules.Notes` module. It is recommended to
        use a namespace for each module that matches the example below:

        .. code-block:: python

            dashboard.register_module_asset({
                'file': 'templates/my-module/js/my-script.js',
                'url': '/module/my-module/js/my-script.js'
            })

        :param asset: A dictionary with keys :code:`'file'` and :code:`'url'`.
        :type asset: dict
        """
        self._module_assets.append(asset)

    def run(self, *args, **kwargs):
        """Run the dashboard webserver.

        Use :py:meth:`~.main` instead of this method for the command-line
        interface. Arguments to this function are passed directly to
        :py:meth:`flask.Flask.run`.
        """
        host = self.config["HOST"]
        port = self.config["PORT"]
        debug = self.config["DEBUG"]
        max_retries = 5

        for _ in range(max_retries):
            try:
                self.app.run(host=host, port=port, debug=debug, *args, **kwargs)
                break
            except OSError as e:
                logger.warning(e)
                if port:
                    port += 1
                pass

    @lru_cache
    def _schema_variables(self):
        schema = self.project.detect_schema(exclude_const=True)
        return [key for key in schema]

    @lru_cache
    def _project_min_len_unique_id(self):
        return self.project.min_len_unique_id()

    def job_title(self, job):
        """Override this method for custom job titles.

        This method generates job titles. By default, the title is a pretty
        (but verbose) form of the job state point, based on the project schema.

        :param job: The job being titled.
        :type job: :py:class:`signac.job.Job`
        :returns: Title to be displayed.
        :rtype: str
        """

        def _format_num(num):
            if isinstance(num, bool):
                return str(num)
            elif isinstance(num, Real):
                return str(round(num, 2))
            return str(num)

        try:
            s = []
            for keys in sorted(self._schema_variables()):
                v = job.statepoint()[keys[0]]
                try:
                    for key in keys[1:]:
                        v = v[key]
                except KeyError:  # Particular key is present in overall
                    continue  # schema, but not this state point.
                else:
                    s.append("{}={}".format(".".join(keys), _format_num(v)))
            return " ".join(s)
        except Exception as error:
            logger.debug(
                "Error while generating job title: '{}'. "
                "Returning job-id as fallback.".format(error)
            )
            return str(job)

    def job_subtitle(self, job):
        """Override this method for custom job subtitles.

        This method generates job subtitles. By default, the subtitle is a
        minimal unique substring of the job id.

        :param job: The job being subtitled.
        :type job: :py:class:`signac.job.Job`
        :returns: Subtitle to be displayed.
        :rtype: str
        """
        return str(job)[: max(8, self._project_min_len_unique_id())]

    def job_sorter(self, job):
        """Override this method for custom job sorting.

        This method returns a key that can be compared to sort jobs. By
        default, the sorting key is based on :py:func:`Dashboard.job_title`,
        with natural sorting of numbers. Good examples of such keys are
        strings or tuples of properties that should be used to sort.

        :param job: The job being sorted.
        :type job: :py:class:`signac.job.Job`
        :returns: Key for sorting.
        :rtype: any comparable type
        """
        key = natsort.natsort_keygen(key=self.job_title, alg=natsort.REAL)
        return key(job)

    @lru_cache
    def _get_all_jobs(self):
        return sorted(self.project.find_jobs(), key=self.job_sorter)

    @lru_cache(maxsize=100)
    def _job_search(self, query):
        if "$where" in query and not self.config.get("ALLOW_WHERE", False):
            flash(
                "Searches using $where allow arbitrary code execution and "
                "are only allowed when the configuration option "
                "'ALLOW_WHERE' is enabled. See also: <a href=\"https://docs.signac.io/projects/dashboard/en/latest/security.html\">Security Guidelines</a>",  # noqa:E501
                "warning",
            )
            raise RuntimeError("ALLOW_WHERE must be enabled for this query.")

        try:
            if query is None:
                f = None
            else:
                try:
                    f = json.loads(query)
                except json.JSONDecodeError:
                    query = shlex.split(query)
                    f = signac.filterparse.parse_filter_arg(query)
                    flash(f"Search string interpreted as '{json.dumps(f)}'.")
            jobs = self.project.find_jobs(filter=f)
            return sorted(jobs, key=lambda job: self.job_sorter(job))
        except json.JSONDecodeError as error:
            flash(
                "Failed to parse query argument. "
                "Ensure that '{}' is valid JSON!".format(query),
                "warning",
            )
            raise error

    @lru_cache(maxsize=65536)
    def _job_details(self, job):
        return {
            "job": job,
            "title": self.job_title(job),
            "subtitle": self.job_subtitle(job),
        }

    def _setup_pagination(self, jobs):
        total_count = len(jobs) if isinstance(jobs, list) else 0
        page = request.args.get("page", 1)
        try:
            page = int(page)
        except (ValueError, TypeError):
            page = 1
            flash(f"Pagination Error. Displaying page {page}.", "danger")
        try:
            per_page = int(request.args.get("per_page", self.config["PER_PAGE"]))
            if per_page < 1:
                raise ValueError("At least one job is required per page.")
        except ValueError:
            flash(
                f"Pagination Error. Displaying {per_page} jobs per page.",
                "danger",
            )
        pagination = Pagination(page, per_page, total_count)
        if pagination.page < 1 or pagination.page > pagination.pages:
            pagination.page = max(1, min(pagination.page, pagination.pages))
            if pagination.pages > 0:
                flash(
                    f"Pagination Error. Displaying page {pagination.page}.",
                    "danger",
                )
        return pagination

    def _setup_enabled_module_indices(self):
        enabled_module_indices = {}
        for context_name, context_modules in self._modules_by_context.items():
            enabled_module_indices[context_name] = [
                i for i, m in enumerate(context_modules) if m.enabled
            ]
        return enabled_module_indices

    def _render_job_view(self, *args, **kwargs):
        g.active_page = "jobs"
        session["context"] = "JobContext"
        session.setdefault(
            "enabled_module_indices", self._setup_enabled_module_indices()
        )
        view_mode = request.args.get("view", kwargs.get("default_view", "list"))
        if view_mode == "grid":
            if (
                len(session.get("enabled_module_indices", {}).get("JobContext", []))
                == 0
            ):
                flash("No modules for the JobContext are enabled.", "info")
            return render_template("jobs_grid.html", *args, **kwargs)
        elif view_mode == "list":
            return render_template("jobs_list.html", *args, **kwargs)
        else:
            return self._render_error(ValueError(f"Invalid view mode: {view_mode}"))

    def _render_project_view(self, *args, **kwargs):
        g.active_page = "project"
        session["context"] = "ProjectContext"
        session.setdefault(
            "enabled_module_indices", self._setup_enabled_module_indices()
        )
        if (
            len(session.get("enabled_module_indices", {}).get("ProjectContext", []))
            == 0
        ):
            flash("No modules for the ProjectContext are enabled.", "info")
        return render_template("project_info.html", *args, **kwargs)

    def _render_error(self, error):
        if isinstance(error, Exception):
            error_string = f"{type(error).__name__}: {error}"
        else:
            error_string = error
        logger.error(error_string)
        flash(error_string, "danger")
        return render_template("error.html")

    def _get_job_details(self, jobs):
        return [self._job_details(job) for job in list(jobs)]

    def add_url(
        self, import_name, url_rules=[], import_file="signac_dashboard", **kwargs
    ):
        r"""Add a route to the dashboard.

        This method allows custom view functions to be triggered for specified
        routes. These view functions are imported lazily, when their route
        is triggered. For example, write a file :code:`my_views.py`:

        .. code-block:: python

            def my_custom_view(dashboard):
                return 'This is a custom message.'

        Then, in :code:`dashboard.py`:

        .. code-block:: python

            from signac_dashboard import Dashboard

            class MyDashboard(Dashboard):
                pass

            if __name__ == '__main__':
                dashboard = MyDashboard()
                dashboard.add_url('my_custom_view', url_rules=['/custom-url'],
                                  import_file='my_views')
                dashboard.main()

        Finally, launching the dashboard with :code:`python dashboard.py run`
        and navigating to :code:`/custom-url` will show the custom
        message. This can be used in conjunction with user-provided jinja
        templates and the method :py:func:`flask.render_template` for extending
        dashboard functionality.

        :param import_name: The view function name to be imported.
        :type import_name: str
        :param url_rules: A list of URL rules, see
            :py:meth:`flask.Flask.add_url_rule`.
        :type url_rules: list
        :param import_file: The module from which to import (default:
            :code:`'signac_dashboard'`).
        :type import_file: str
        :param \\**kwargs: Additional keyword arguments to pass to
            :py:meth:`flask.Flask.add_url_rule`.
        """
        if import_file is not None:
            import_name = import_file + "." + import_name
        view = LazyView(dashboard=self, import_name=import_name)
        for url_rule in url_rules:
            self.app.add_url_rule(
                rule=url_rule,
                view_func=view,
                **kwargs,
            )

    def _register_routes(self):
        """Register routes with the Flask application.

        This method configures context processors, templates, and sets up
        routes for a basic Dashboard instance. Additionally, routes declared by
        modules are registered by this method.
        """
        dashboard = self

        @dashboard.app.after_request
        def prevent_caching(response):
            if "Cache-Control" not in response.headers:
                response.headers["Cache-Control"] = "no-store"
            return response

        @dashboard.app.context_processor
        def injections():
            # inject new variables into the template context
            session.setdefault(
                "enabled_module_indices", self._setup_enabled_module_indices()
            )
            return {
                "APP_NAME": "signac-dashboard",
                "APP_VERSION": __version__,
                "CARDS_PER_ROW": self.config["CARDS_PER_ROW"],
                "modules": self.modules,
                "modules_by_context": self._modules_by_context,
                "enabled_module_indices": session["enabled_module_indices"],
                "module_assets": self._module_assets,
            }

        # Add pagination support from http://flask.pocoo.org/snippets/44/
        @dashboard.app.template_global()
        def url_for_other_page(page):
            args = request.args.copy()
            args["page"] = page
            return url_for(request.endpoint, **args)

        @dashboard.app.template_global()
        def modify_query(**new_values):
            args = request.args.copy()
            for key, value in new_values.items():
                args[key] = value
            return f"{request.path}?{url_encode(args)}"

        @dashboard.app.errorhandler(404)
        def page_not_found(error):
            return self._render_error(str(error))

        @dashboard.login_manager.unauthorized_handler
        def unauthorized_handler():
            request_url = request.url
            if "module" not in request_url:
                session["redirect_url"] = request_url
            return render_template("login.html")

        @dashboard.app.route("/login", methods=["GET", "POST"])
        def login():
            redirect_url = session.pop("redirect_url", "/")
            if flask_login.current_user.is_authenticated:
                # in case the user goes to the login page via browser history
                return redirect(redirect_url)

            if request.method == "POST":
                provided_token = request.form.get("token")
            else:
                provided_token = request.args.get("token")  # None if not given

            if provided_token == self.config["ACCESS_TOKEN"]:
                # Log the user in and redirect to the previous page (if applicable)
                user = User(provided_token)
                flask_login.login_user(user)
                return redirect(redirect_url)

            elif provided_token is None:
                # First time visiting page, so don't display error.
                return render_template("login.html")
            else:
                flash("Incorrect token", "danger")
                if request.method == "GET":
                    return redirect("/login")
                elif request.method == "POST":
                    if redirect_url == "/":
                        redirect_url = "/login"
                    return redirect(redirect_url)

        @dashboard.app.route("/logout")
        def logout():
            flask_login.logout_user()
            return redirect(url_for("login"))

        @dashboard.app.route("/favicon.ico")
        @flask_login.login_required
        def favicon():
            return url_for("static", filename="favicon.ico")

        # These routes are protected within the LazyView utility class
        self.add_url("views.home", ["/"])
        self.add_url("views.settings", ["/settings"])
        self.add_url("views.search", ["/search"])
        self.add_url("views.project_info", ["/project/"])
        self.add_url("views.jobs_list", ["/jobs/"])
        self.add_url("views.show_job", ["/jobs/<jobid>"])
        self.add_url(
            "views.get_file",
            ["/jobs/<jobid>/file/<path:filename>", "/project/file/<path:filename>"],
        )
        self.add_url("views.change_modules", ["/modules"], methods=["POST"])

    def update_cache(self):
        """Clear project and dashboard server caches.

        The dashboard relies on caching for performance. If the data space is
        altered, this method may need to be called before the dashboard
        reflects those changes.
        """
        # Try to update signac project cache. Requires signac 0.9.2 or later.
        with warnings.catch_warnings():
            warnings.simplefilter(action="ignore", category=FutureWarning)
            try:
                self.project.update_cache()
            except Exception:
                pass

        # Clear caches of all dashboard methods
        members = inspect.getmembers(self, predicate=inspect.ismethod)
        for func in filter(
            lambda f: hasattr(f, "cache_clear"), map(lambda x: x[1], members)
        ):
            func.cache_clear()

    def __call__(self, environ, start_response):
        """Call the dashboard as a WSGI application."""
        return self.app(environ, start_response)

    def main(self, command_args=None):
        """Run the command line interface.

        Call this function to use signac-dashboard from its command line
        interface. For example, save this script as :code:`dashboard.py`:

        .. code-block:: python

            from signac_dashboard import Dashboard

            class MyDashboard(Dashboard):
                pass

            if __name__ == '__main__':
                MyDashboard().main()

        Then the dashboard can be launched with:

        .. code-block:: bash

            python dashboard.py run

        :param command_args: List of CLI arguments to pass, e.g.
            ``["--debug", "--port", "8889"]`` (default: None).
        :type command_args: list
        """
        if command_args is not None and len(command_args) == 0:
            command_args = None

        def _run(args):
            kwargs = vars(args)
            if kwargs.get("host", None) is not None:
                self.config["HOST"] = kwargs.pop("host")
            if kwargs.get("port", None) is not None:
                self.config["PORT"] = kwargs.pop("port")
            self.config["PROFILE"] = kwargs.pop("profile")
            self.config["DEBUG"] = kwargs.pop("debug")

            if self.config["ACCESS_TOKEN"] is not None:
                print(
                    f"To access this server, connect to:\n\n"
                    f"http://{self.config['HOST']}:{self.config['PORT']}/"
                    f"login?token={self.config['ACCESS_TOKEN']}\n"
                )

            self.run()

        parser = argparse.ArgumentParser(
            description="signac-dashboard is a web-based data visualization "
            "and analysis tool, part of the signac framework."
        )
        parser.add_argument(
            "--debug",
            action="store_true",
            help="Show traceback on error for debugging.",
        )
        parser.add_argument(
            "--version",
            action="store_true",
            help="Display the version number and exit.",
        )
        subparsers = parser.add_subparsers()

        parser_run = subparsers.add_parser("run")
        parser_run.add_argument(
            "-p",
            "--profile",
            action="store_true",
            help="Enable flask performance profiling.",
        )
        parser_run.add_argument(
            "-d", "--debug", action="store_true", help="Enable flask debug mode."
        )
        parser_run.add_argument(
            "--host", type=str, help="Host (binding address). Default: localhost"
        )
        parser_run.add_argument(
            "--port", type=int, help="Port to listen on. Default: 8888"
        )
        parser_run.set_defaults(func=_run)

        # This is a hack, as argparse itself does not
        # allow to parse only --version without any
        # of the other required arguments.
        if "--version" in sys.argv:
            print("signac-dashboard", __version__)
            sys.exit(0)

        args = parser.parse_args(command_args)

        if args.debug:
            logger.setLevel(logging.DEBUG)

        if not hasattr(args, "func"):
            parser.print_usage()
            sys.exit(2)
        try:
            self.observer.start()
            args.func(args)
        except RuntimeWarning as warning:
            logger.warning(f"Warning: {warning}")
            if args.debug:
                raise
            sys.exit(1)
        except Exception as error:
            logger.error(f"Error: {error}")
            if args.debug:
                raise
            sys.exit(1)
        finally:
            self.observer.stop()
            self.observer.join()



# Copyright (c) 2019 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
from math import ceil


class Pagination:
    """Paginate a set of items.

    Adapted from http://flask.pocoo.org/snippets/44/.

    :param int page: Current page number.
    :param int per_page: Number of items per page. If 0 or `None`, all items
                         will be displayed on one page.
    :param int total_count: The total number of items being paginated.
    """

    def __init__(self, page, per_page, total_count):
        self.page = page
        self.per_page = per_page
        self.total_count = total_count

    @property
    def pages(self):
        if self.per_page is None or self.per_page == 0:
            return 1
        else:
            return int(ceil(self.total_count / float(self.per_page)))

    @property
    def has_prev(self):
        return self.page > 1

    @property
    def has_next(self):
        return self.page < self.pages

    @property
    def first_item(self):
        if self.per_page is None:
            return 0
        else:
            return max((self.page - 1) * self.per_page, 0)

    @property
    def last_item(self):
        if self.per_page is None:
            return self.total_count
        else:
            return min(self.page * self.per_page, self.total_count)

    def paginate(self, items):
        if items is None:
            return []
        else:
            return items[self.first_item : self.last_item]

    def item_counts(self, tag="jobs"):
        if self.total_count > 0:
            return "{} to {} of {} {}".format(
                self.first_item + 1, self.last_item, self.total_count, tag
            )
        else:
            return f"{self.total_count} {tag}"

    def iter_pages(self, left_edge=2, left_current=3, right_current=5, right_edge=2):
        last = 0
        for num in range(1, self.pages + 1):
            if (
                num <= left_edge
                or (self.page - left_current - 1 < num < self.page + right_current)
                or num > self.pages - right_edge
            ):
                if last + 1 != num:
                    yield None
                yield num
                last = num



# Copyright (c) 2022 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
from typing import Set

from flask import render_template


class Module:
    """Base class for modules, the building blocks of the dashboard.

    A module turns select information from the job directory or project
    directory into cards displayed on the dashboard. The cards pull content from
    the job directory or project directory depending on whether the module
    ``context`` is ``'JobContext'`` or ``'ProjectContext'``.

    :param name: Name of the module for card titles.
    :type name: str
    :param context: Context in which this module's cards will be displayed, either
        :code:`'JobContext'` *or* :code:`'ProjectContext'`.
    :type context: str
    :param template: Path to a template file for this module's cards (e.g.
        :code:`cards/my_module.html`, without the template directory prefix
        :code:`templates/`).
    :type template: str
    :param enabled: Whether the module's cards will be displayed. (default: :code:`True`)
    :type enabled: bool

    **Custom modules:** User-defined module classes should be a subclass of
    :py:class:`~.Module` and define the function :py:meth:`~.Module.get_cards`.
    Template files are written in HTML/Jinja-compatible syntax.
    See `this example <https://github.com/glotzerlab/signac-dashboard/tree/main/examples/custom-modules>`_.

    **Module assets:** If a module requires scripts or stylesheets to be
    included for its content to be rendered, they must be handled by the
    callback :py:meth:`.register`. For example, a module inheriting from
    the base :py:class:`signac_dashboard.Module` class may implement this by
    overriding the default method as follows:

    .. code-block:: python

        def register(self, dashboard):
            assets = ['js/my-script.js', 'css/my-style.css']
            for asset in assets:
                dashboard.register_module_asset({
                    'file': 'templates/my-module/{}'.format(asset),
                    'url': '/module/my-module/{}'.format(asset)
                })

    Then, when the module is active, its assets will be included and a
    route will be created that returns the asset file.

    **Module routes:** The callback :py:meth:`.register` allows modules
    to implement custom routes, such as methods that should be triggered by
    :code:`POST` requests or custom APIs. For example, a module inheriting from
    the base :py:class:`signac_dashboard.Module` class may implement this by
    overriding the default method as follows:

    .. code-block:: python

        def register(self, dashboard):
            @dashboard.app.route('/module/my-module/update', methods=['POST'])
            @flask_login.login_required
            def my_module_update():
                # Perform update
                return "Saved."

    """  # noqa: E501

    _supported_contexts: Set[str] = set()

    def __init__(self, name, context, template, enabled=True):
        self._module = self.__module__
        self._moduletype = self.__class__.__name__
        self.name = name
        if len(self._supported_contexts) == 0:
            raise ValueError(f"{self._moduletype} is not supported by any contexts.")
        if context not in self._supported_contexts:
            raise RuntimeError(
                f"{self._moduletype} does not support the {context}, only "
                f"{self._supported_contexts}."
            )
        self.context = context
        self.template = template
        self.enabled = enabled

    def get_cards(self):
        """Return this module's cards for rendering.

        The cards are returned as a list of dictionaries with keys
        :code:`'name'` and :code:`'content'`.

        :returns: List of module cards.
        :rtype: list
        """
        return [{"name": self.name, "content": render_template(self.template)}]

    def enable(self):
        """Enable this module."""
        self.enabled = True

    def disable(self):
        """Disable this module."""
        self.enabled = False

    def toggle(self):
        """Toggle this module."""
        self.enabled = not self.enabled

    def register(self, dashboard):
        """Register this module with the dashboard.

        This method is a callback used to register assets and routes, as well
        as any other initialization that accesses or modifies the dashboard.

        :param dashboard: The dashboard invoking this callback method.
        :type dashboard: :py:class:`signac_dashboard.Dashboard`
        """
        pass



from flask import render_template, url_for
from signac._utility import _to_hashable

from signac_dashboard.module import Module
from signac_dashboard.util import abbr_value


class _DictPlaceholder:
    pass


class Navigator(Module):
    """Displays links to jobs differing in one state point parameter.

    This module uses the project schema to determine which state points vary, then displays links to
    jobs with adjacent values of these parameters in a table. Schema detection can be slow on slow
    file systems, so this module caches the project schema. Therefore, this module may not update if
    the signac project changes while the signac-dashboard is running.

    :param context: Supports :code:`'JobContext'`
    :type context: str
    :param max_chars: Truncation length of state point values (default: 6).
    :type max_chars: int
    """

    _supported_contexts = {"JobContext"}

    def __init__(
        self,
        name="Navigator",
        context="JobContext",
        template="cards/navigator.html",
        max_chars=6,
        **kwargs,
    ):
        super().__init__(name=name, context=context, template=template, **kwargs)
        self.max_chars = max_chars

    def _link_label(self, job, project, key, other_val):
        """Return the url and label for the job with job.sp[key] == other_val."""
        similar_statepoint = job.statepoint()  # modifiable
        similar_statepoint.update({key: other_val})

        # Look only for exact matches that result from only changing one parameter
        # in case of heterogeneous schema
        other_job = project.open_job(similar_statepoint)
        if other_job in project:
            link = url_for("show_job", jobid=other_job.id)
            label = abbr_value(other_val, self.max_chars)
        else:
            link = None
            label = f"no match for {other_val}"
        return link, label

    def get_cards(self, job):
        project = self._dashboard_project

        nearby_jobs = {}
        sp_copy = job.sp()

        # for each parameter in the schema, find the next and previous jobs and get links to them
        for key, schema_values in self._sorted_schema.items():
            # allow comparison with output of schema, which is hashable
            value = _to_hashable(sp_copy.get(key, _DictPlaceholder))
            if value is _DictPlaceholder:
                # Possible if schema is heterogeneous
                continue
            value_index = schema_values.index(value)

            query_index = value_index - 1
            while query_index >= 0:
                prev_val = schema_values[query_index]
                link, label = self._link_label(job, project, key, prev_val)
                if link is None:
                    query_index -= 1
                else:
                    break
            else:
                link = None
                label = "min"
            previous_label = (link, label)

            query_index = value_index + 1
            while query_index <= len(schema_values) - 1:
                next_val = schema_values[query_index]
                link, label = self._link_label(job, project, key, next_val)
                if link is None:
                    query_index += 1
                else:
                    break
            else:
                link = None
                label = "max"
            next_label = (link, label)

            if previous_label[0] is not None or next_label[0] is not None:
                nearby_jobs[key] = (
                    abbr_value(value, self.max_chars),
                    (previous_label, next_label),
                )

        return [
            {
                "name": self.name,
                "content": render_template(self.template, job_nav=nearby_jobs.items()),
            }
        ]

    def register(self, dashboard):
        """Sorts and caches non-constant schema schema_values."""
        self._dashboard_project = dashboard.project

        # Tell user because this can take a long time
        print("Detecting project schema for Navigator...", end="", flush=True)
        schema = dashboard.project.detect_schema(exclude_const=True)
        print("done.")

        # turn dict of sets of lists ...into list of parameters
        sorted_schema = {}
        for key, project_values in schema.items():
            this_key_vals = set()
            for typename in project_values.keys():
                this_key_vals.update(project_values[typename])
            try:
                sorted_schema[key] = sorted(this_key_vals)
            except TypeError:
                # cannot sort between different types, so leave in order
                sorted_schema[key] = list(this_key_vals)

        self._sorted_schema = dict(sorted(sorted_schema.items(), key=lambda t: t[0]))



# Copyright (c) 2022 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
from flask import Markup, render_template

from signac_dashboard.module import Module

try:
    import markdown

    MARKDOWN = True
except ImportError:
    MARKDOWN = False


class TextDisplay(Module):
    """Render custom text or Markdown in a card.

    This module calls a user-provided function to display text or Markdown
    content in a card. Rendering Markdown requires the :code:`markdown`
    library to be installed. Example:

    .. code-block:: python

        from signac_dashboard.modules import TextDisplay

        def my_text(job):
            return f"This job id is {job}."

        modules = [TextDisplay(message=my_text)]

    :param context: Supports :code:`'JobContext'` and :code:`'ProjectContext'`.
    :type context: str
    :param message: A callable accepting one argument of type
        :py:class:`signac.job.Job` or :py:class:`signac.Project`
        and returning text or Markdown content.
    :type message: callable
    :param markdown: Enables Markdown rendering if True (default: :code:`False`).
    :type markdown: bool
    """

    _supported_contexts = {"JobContext", "ProjectContext"}

    def __init__(
        self,
        name="Text Display",
        context="JobContext",
        message=lambda job_or_project: "No message provided.",
        markdown=False,
        **kwargs,
    ):
        super().__init__(
            name=name,
            context=context,
            template="cards/text_display.html",
            **kwargs,
        )
        self.message = message
        self.markdown = markdown

    def get_cards(self, job_or_project):
        msg = self.message(job_or_project)
        if self.markdown:
            if MARKDOWN:
                msg = Markup(
                    markdown.markdown(msg, extensions=["markdown.extensions.attr_list"])
                )
            else:
                msg = "Error: Install the 'markdown' library to render Markdown."
        return [{"name": self.name, "content": render_template(self.template, msg=msg)}]



# Copyright (c) 2022 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
from collections import OrderedDict

from flask import render_template

from signac_dashboard.module import Module


class StatepointList(Module):
    """Displays the job state point.

    :param context: Supports :code:`'JobContext'`.
    :type context: str
    """

    _supported_contexts = {"JobContext"}

    def __init__(
        self,
        name="Statepoint Parameters",
        context="JobContext",
        template="cards/statepoint_list.html",
        **kwargs,
    ):
        super().__init__(
            name=name,
            context=context,
            template=template,
            **kwargs,
        )

    def get_cards(self, job):
        sp = OrderedDict(sorted(job.statepoint().items(), key=lambda t: t[0]))
        return [
            {
                "name": self.name,
                "content": render_template(self.template, statepoint=sp),
            }
        ]



# Copyright (c) 2022 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
from flask import render_template

from signac_dashboard.module import Module
from signac_dashboard.util import escape_truncated_values


class Schema(Module):
    """Displays the project schema.

    Long values can be optionally truncated.

    :param context: Supports :code:`'ProjectContext'`.
    :type context: str
    :param max_chars: Truncation length for schema values (default:
        :code:`None`).
    :type max_chars: int
    :param exclude_const: Exclude all state point parameters that are shared by all jobs
        in this project (default: :code:`False`).
    :type exclude_const: bool
    :param subset: A sequence of jobs or job ids specifying a subset over which the state
        point schema should be detected (default: :code:`None`).
    """

    _supported_contexts = {"ProjectContext"}

    def __init__(
        self,
        name="Project Schema",
        context="ProjectContext",
        template="cards/escaped_dict_display.html",
        max_chars=None,
        exclude_const=False,
        subset=None,
        **kwargs,
    ):
        super().__init__(
            name=name,
            context=context,
            template=template,
            **kwargs,
        )
        self.max_chars = max_chars
        self.exclude_const = exclude_const
        self.subset = subset

    def get_cards(self, project):
        schema = project.detect_schema(
            exclude_const=self.exclude_const, subset=self.subset
        )
        schema = dict(schema.items())

        # We manually escape the schema contents since the field is marked
        # "safe" in the Jinja template. This is necessary because we added
        # custom HTML for "[Truncated]" fields
        schema = escape_truncated_values(schema, self.max_chars)

        return [
            {
                "name": self.name,
                "content": render_template(self.template, escaped_dict=schema),
            }
        ]



# Copyright (c) 2022 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
from . import modules
from .dashboard import Dashboard
from .module import Module
from .version import __version__

__all__ = [
    "__version__",
    "Dashboard",
    "Module",
    "modules",
]



# Copyright (c) 2022 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
import os

from flask import render_template

from signac_dashboard.module import Module


class FileList(Module):
    """Lists files in the job directory with download links.

    :param context: Supports :code:`'JobContext'`.
    :type context: str
    :param prefix_jobid: Whether filenames should be prefixed with the job id
        when being downloaded (default: :code:`True`).
    :type prefix_jobid: bool
    """

    _supported_contexts = {"JobContext"}

    def __init__(
        self,
        name="File List",
        context="JobContext",
        template="cards/file_list.html",
        prefix_jobid=True,
        **kwargs,
    ):
        super().__init__(
            name=name,
            context=context,
            template=template,
            **kwargs,
        )
        self.prefix_jobid = prefix_jobid

    def download_name(self, job, filename):
        if self.prefix_jobid:
            return f"{str(job)}_{filename}"
        else:
            return filename

    def get_cards(self, job):
        files = sorted(
            (
                {
                    "name": filename,
                    "jobid": job._id,
                    "download": self.download_name(job, filename),
                }
                for filename in os.listdir(job.path)
            ),
            key=lambda filedata: filedata["name"],
        )
        return [
            {"name": self.name, "content": render_template(self.template, files=files)}
        ]
